package com.terraformersmc.terraform.biome.builder;

import net.minecraft.class_1299;
import net.minecraft.class_1959;
import net.minecraft.class_2680;
import net.minecraft.class_2893;
import net.minecraft.class_2975;
import net.minecraft.class_3031;
import net.minecraft.class_3037;
import net.minecraft.class_3111;
import net.minecraft.class_3195;
import net.minecraft.class_3273;
import net.minecraft.class_3276;
import net.minecraft.class_3284;
import net.minecraft.class_3297;
import net.minecraft.class_3504;
import net.minecraft.class_3523;
import net.minecraft.class_3531;
import net.minecraft.class_4632;
import net.minecraft.class_4633;
import net.minecraft.class_4638;
import net.minecraft.class_4656;
import net.minecraft.class_4657;
import net.minecraft.world.gen.feature.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

public class TerraformBiome extends class_1959 {
	private int grassColor;
	private int foliageColor;

	private TerraformBiome(class_1959.class_1960 biomeSettings, ArrayList<class_1964> spawns) {
		super(biomeSettings);
		for (class_1964 entry : spawns) {
			this.method_8708(entry.field_9389.method_5891(), entry);
		}
	}

	public void setGrassAndFoliageColors(int grassColor, int foliageColor) {
		this.grassColor = grassColor;
		this.foliageColor = foliageColor;
	}

	@Override
	public int method_8711(double x, double y) {
		if (grassColor == -1) {
			return super.method_8711(x, y);
		}

		return grassColor;
	}

	@Override
	public int method_8698() {
		if (foliageColor == -1) {
			return super.method_8698();
		}

		return foliageColor;
	}

	public static TerraformBiome.Builder builder() {
		return new Builder();
	}

	public static final class Builder extends BuilderBiomeSettings {

		private ArrayList<DefaultFeature> defaultFeatures = new ArrayList<>();
		private ArrayList<FeatureEntry> features = new ArrayList<>();
		private Map<class_3195<class_3037>, class_3037> structureFeatures = new HashMap<>();
		private Map<class_2975, Integer> treeFeatures = new HashMap<>();
		private Map<class_2975, Integer> rareTreeFeatures = new HashMap<>();
		private Map<class_2680, Integer> plantFeatures = new HashMap<>();
		private Map<class_2680, Integer> doublePlantFeatures = new HashMap<>();
		private ArrayList<class_1964> spawnEntries = new ArrayList<>();
		private int grassColor = -1;
		private int foliageColor = -1;
		private boolean template = false;
		// NOTE: Make sure to add any additional fields to the Template copy code down below!

		Builder() {
			super();

			parent(null);
		}

		Builder(Builder existing) { // Template copy code
			super(existing);

			this.defaultFeatures.addAll(existing.defaultFeatures);
			this.features.addAll(existing.features);
			this.structureFeatures.putAll(existing.structureFeatures);
			this.treeFeatures.putAll(existing.treeFeatures);
			this.rareTreeFeatures.putAll(existing.rareTreeFeatures);
			this.plantFeatures.putAll(existing.plantFeatures);
			this.doublePlantFeatures.putAll(existing.doublePlantFeatures);
			this.spawnEntries.addAll(existing.spawnEntries);

			this.grassColor = existing.grassColor;
			this.foliageColor = existing.foliageColor;
		}

		@SuppressWarnings("unchecked")
		public class_1959 build() {
			if (template) {
				throw new IllegalStateException("Tried to call build() on a frozen Builder instance!");
			}

			// Add SpawnEntries
			TerraformBiome biome = new TerraformBiome(this, this.spawnEntries);

			// Set grass and foliage colors
			biome.setGrassAndFoliageColors(this.grassColor, this.foliageColor);

			// Add structures
			for (Map.Entry<class_3195<class_3037>, class_3037> structure : structureFeatures.entrySet()) {
				biome.method_8710(structure.getKey().method_23397(structure.getValue()));
			}

			// Tree Feature stuff
			if (treeFeatures.size() > 0) {

				// Determine the total tree count

				int totalTreesPerChunk = 0;
				for (Integer count : treeFeatures.values()) {
					totalTreesPerChunk += count;
				}

				// Add each tree

				for (Map.Entry<class_2975, Integer> tree : treeFeatures.entrySet()) {
					class_2975 feature = tree.getKey();
					int count = tree.getValue();

					float weight = (float) count / totalTreesPerChunk;

					biome.method_8719(
						class_2893.class_2895.field_13178,
						feature.method_23388(
							class_3284.field_14267.method_23475(new class_3276(count, 0.1F * weight, 1)))
					);
				}
			}

			// Rare tree features

			for (Map.Entry<class_2975, Integer> tree : rareTreeFeatures.entrySet()) {
				class_2975 feature = tree.getKey();
				int chance = tree.getValue();

				biome.method_8719(
					class_2893.class_2895.field_13178,
					feature.method_23388(class_3284.field_14259.method_23475(new class_3297(chance)))
				);
			}

			// Add any minecraft (default) features

			for (DefaultFeature defaultFeature : defaultFeatures) {
				defaultFeature.add(biome);
			}

			// Add custom features that don't fit in the templates

			for (FeatureEntry feature : features) {
				biome.method_8719(feature.getStep(), feature.getFeature());
			}

			// Add Plant decoration features

			class_4657 weightedStateProvider = new class_4657();
			int chanceTotal = 0;
			for (Map.Entry<class_2680, Integer> plant : plantFeatures.entrySet()) {
				weightedStateProvider.method_23458(plant.getKey(), plant.getValue());
				chanceTotal += plant.getValue();
			}
			biome.method_8719(
				class_2893.class_2895.field_13178,
				class_3031.field_21220.method_23397(
					new class_4638.class_4639(weightedStateProvider, new class_4633())
						.method_23417(32)
						.method_23424())
					.method_23388(
						class_3284.field_14240.method_23475(new class_3273(chanceTotal))));

			// Add Double Plant decoration features

			for (Map.Entry<class_2680, Integer> doublePlant : doublePlantFeatures.entrySet()) {
				biome.method_8719(
					class_2893.class_2895.field_13178,
					class_3031.field_21220.method_23397(
						new class_4638.class_4639(new class_4656(doublePlant.getKey()), new class_4632())
							.method_23417(64)
							.method_23419()
							.method_23424())
						.method_23388(
							class_3284.field_14253.method_23475(new class_3273(doublePlant.getValue()))));
			}

			return biome;
		}

		@Override
		public <SC extends class_3531> Builder configureSurfaceBuilder(class_3523<SC> builder, SC config) {
			super.configureSurfaceBuilder(builder, config);

			return this;
		}

		@Override
		public Builder surfaceBuilder(class_3504<?> surfaceBuilder) {
			super.surfaceBuilder(surfaceBuilder);

			return this;
		}

		@Override
		public Builder precipitation(class_1959.class_1963 precipitation) {
			super.precipitation(precipitation);

			return this;
		}

		@Override
		public Builder category(class_1959.class_1961 category) {
			super.category(category);

			return this;
		}

		@Override
		public Builder depth(float depth) {
			super.depth(depth);

			return this;
		}

		@Override
		public Builder scale(float scale) {
			super.scale(scale);

			return this;
		}

		@Override
		public Builder temperature(float temperature) {
			super.temperature(temperature);

			return this;
		}

		@Override
		public Builder downfall(float downfall) {
			super.downfall(downfall);

			return this;
		}

		@Override
		public Builder waterColor(int color) {
			super.waterColor(color);

			return this;
		}

		@Override
		public Builder waterFogColor(int color) {
			super.waterFogColor(color);

			return this;
		}

		@Override
		public Builder parent(String parent) {
			super.parent(parent);

			return this;
		}

		public TerraformBiome.Builder addTreeFeature(class_2975 feature, int numPerChunk) {
			this.treeFeatures.put(feature, numPerChunk);
			return this;
		}

		public TerraformBiome.Builder addRareTreeFeature(class_2975 feature, int chance) {
			this.rareTreeFeatures.put(feature, chance);
			return this;
		}

		public TerraformBiome.Builder addGrassFeature(class_2680 blockState, int count) {
			this.plantFeatures.put(blockState, count);
			return this;
		}

		public TerraformBiome.Builder addDoubleGrassFeature(class_2680 blockState, int count) {
			this.doublePlantFeatures.put(blockState, count);
			return this;
		}

		public TerraformBiome.Builder addCustomFeature(class_2893.class_2895 step, class_2975 feature) {
			this.features.add(new FeatureEntry(step, feature));
			return this;
		}

		public TerraformBiome.Builder addSpawnEntry(class_1964 entry) {
			this.spawnEntries.add(entry);
			return this;
		}

		public TerraformBiome.Builder addStructureFeature(class_3195<class_3111> feature) {
			return this.addStructureFeature(feature, class_3037.field_13603);
		}

		public <FC extends class_3037> TerraformBiome.Builder addStructureFeature(class_3195<FC> feature, FC config) {
			this.structureFeatures.put((class_3195) feature, config);
			return this;
		}

		public TerraformBiome.Builder addStructureFeatures(class_3195<class_3111>... defaultStructureFeatures) {
			for (class_3195<class_3111> feature : defaultStructureFeatures) {
				this.structureFeatures.put((class_3195) feature, class_3037.field_13603);
			}
			return this;
		}

		public TerraformBiome.Builder addDefaultFeature(DefaultFeature feature) {
			defaultFeatures.add(feature);
			return this;
		}

		public TerraformBiome.Builder addDefaultFeatures(DefaultFeature... features) {
			defaultFeatures.addAll(Arrays.asList(features));
			return this;
		}

		public TerraformBiome.Builder grassColor(int color) {
			grassColor = color;
			return this;
		}

		public TerraformBiome.Builder foliageColor(int color) {
			foliageColor = color;
			return this;
		}

		public TerraformBiome.Builder addDefaultSpawnEntries() {
			this.addSpawnEntry(new class_1959.class_1964(class_1299.field_6115, 12, 4, 4))
				.addSpawnEntry(new class_1959.class_1964(class_1299.field_6093, 10, 4, 4))
				.addSpawnEntry(new class_1959.class_1964(class_1299.field_6132, 10, 4, 4))
				.addSpawnEntry(new class_1959.class_1964(class_1299.field_6085, 8, 4, 4))
				.addSpawnEntry(new class_1959.class_1964(class_1299.field_6108, 10, 8, 8))
				.addSpawnEntry(new class_1959.class_1964(class_1299.field_6079, 100, 4, 4))
				.addSpawnEntry(new class_1959.class_1964(class_1299.field_6051, 95, 4, 4))
				.addSpawnEntry(new class_1959.class_1964(class_1299.field_6054, 5, 1, 1))
				.addSpawnEntry(new class_1959.class_1964(class_1299.field_6137, 100, 4, 4))
				.addSpawnEntry(new class_1959.class_1964(class_1299.field_6046, 100, 4, 4))
				.addSpawnEntry(new class_1959.class_1964(class_1299.field_6069, 100, 4, 4))
				.addSpawnEntry(new class_1959.class_1964(class_1299.field_6091, 10, 1, 4))
				.addSpawnEntry(new class_1959.class_1964(class_1299.field_6145, 5, 1, 1));
			return this;
		}
	}

	public static final class Template {
		private final Builder builder;

		public Template(Builder builder) {
			this.builder = builder;
			builder.template = true;
		}

		public Builder builder() {
			return new Builder(this.builder);
		}
	}
}
