package com.terraformersmc.terraform.block;

import it.unimi.dsi.fastutil.objects.Object2IntMap;
import it.unimi.dsi.fastutil.objects.Object2IntOpenHashMap;
import net.minecraft.block.*;
import net.minecraft.class_10;
import net.minecraft.class_1268;
import net.minecraft.class_1269;
import net.minecraft.class_1299;
import net.minecraft.class_1309;
import net.minecraft.class_1657;
import net.minecraft.class_1750;
import net.minecraft.class_1766;
import net.minecraft.class_1799;
import net.minecraft.class_1922;
import net.minecraft.class_1936;
import net.minecraft.class_1937;
import net.minecraft.class_2246;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2415;
import net.minecraft.class_2470;
import net.minecraft.class_259;
import net.minecraft.class_265;
import net.minecraft.class_2680;
import net.minecraft.class_2689;
import net.minecraft.class_2746;
import net.minecraft.class_3417;
import net.minecraft.class_3419;
import net.minecraft.class_3610;
import net.minecraft.class_3612;
import net.minecraft.class_3726;
import net.minecraft.class_3737;
import net.minecraft.class_3965;
import net.minecraft.class_4538;
import java.util.function.Supplier;

// Rather complex: combine the function of logs, and cobblestone walls.

/**
 * A very complex smaller log block that can connect on all 6 axes and can be waterlogged.
 * Used for things like the Saguaro Cactus.
 */
public class BareSmallLogBlock extends class_2248 implements class_3737 {
	public static final class_2746 UP = class_2746.method_11825("up");
	public static final class_2746 DOWN = class_2746.method_11825("down");
	public static final class_2746 NORTH = class_2746.method_11825("north");
	public static final class_2746 EAST = class_2746.method_11825("east");
	public static final class_2746 SOUTH = class_2746.method_11825("south");
	public static final class_2746 WEST = class_2746.method_11825("west");
	public static final class_2746 WATERLOGGED = class_2746.method_11825("waterlogged");

	private static final int UP_MASK = 1 << class_2350.field_11036.ordinal();
	private static final int DOWN_MASK = 1 << class_2350.field_11033.ordinal();
	private static final int NORTH_MASK = 1 << class_2350.field_11043.ordinal();
	private static final int EAST_MASK = 1 << class_2350.field_11034.ordinal();
	private static final int SOUTH_MASK = 1 << class_2350.field_11035.ordinal();
	private static final int WEST_MASK = 1 << class_2350.field_11039.ordinal();

	protected final class_265[] collisionShapes;
	protected final class_265[] boundingShapes;
	private final Object2IntMap<class_2680> SHAPE_INDEX_CACHE = new Object2IntOpenHashMap<>();

	private final Supplier<class_2248> stripped;

	public BareSmallLogBlock(Supplier<class_2248> stripped, class_2248.class_2251 settings) {
		super(settings);
		this.method_9590(this.method_9595().method_11664()
			.method_11657(UP, false)
			.method_11657(DOWN, false)
			.method_11657(WEST, false)
			.method_11657(EAST, false)
			.method_11657(NORTH, false)
			.method_11657(SOUTH, false)
			.method_11657(WATERLOGGED, false)
		);

		this.collisionShapes = this.createShapes(5);
		this.boundingShapes = this.createShapes(5);
		this.stripped = stripped;
	}

	private int getShapeIndex(class_2680 requested) {
		return this.SHAPE_INDEX_CACHE.computeIntIfAbsent(requested, state -> {
			int mask = 0;

			if (state.method_11654(UP)) {
				mask |= UP_MASK;
			}

			if (state.method_11654(DOWN)) {
				mask |= DOWN_MASK;
			}

			if (state.method_11654(NORTH)) {
				mask |= NORTH_MASK;
			}

			if (state.method_11654(EAST)) {
				mask |= EAST_MASK;
			}

			if (state.method_11654(SOUTH)) {
				mask |= SOUTH_MASK;
			}

			if (state.method_11654(WEST)) {
				mask |= WEST_MASK;
			}

			return mask;
		});
	}

	public class_265[] createShapes(double radius) {
		double lower = 8.0 - radius;
		double upper = 8.0 + radius;

		class_265 center = class_2248.method_9541(lower, lower, lower, upper, upper, upper);

		class_265 down = class_2248.method_9541(lower, 0.0, lower, upper, lower, upper);
		class_265 up = class_2248.method_9541(lower, upper, lower, upper, 16.0, upper);

		// Minus Z: North
		class_265 north = class_2248.method_9541(lower, lower, 0.0, upper, upper, lower);
		class_265 south = class_2248.method_9541(lower, lower, upper, upper, upper, 16.0);

		// Minus X: West
		class_265 west = class_2248.method_9541(0.0, lower, lower, lower, upper, upper);
		class_265 east = class_2248.method_9541(upper, lower, lower, 16.0, upper, upper);

		class_265[] shapes = new class_265[64];

		for (int i = 0; i < 64; i++) {
			class_265 shape = center;

			if ((i & DOWN_MASK) != 0) {
				shape = class_259.method_1084(shape, down);
			}

			if ((i & UP_MASK) != 0) {
				shape = class_259.method_1084(shape, up);
			}

			if ((i & NORTH_MASK) != 0) {
				shape = class_259.method_1084(shape, north);
			}

			if ((i & SOUTH_MASK) != 0) {
				shape = class_259.method_1084(shape, south);
			}

			if ((i & WEST_MASK) != 0) {
				shape = class_259.method_1084(shape, west);
			}

			if ((i & EAST_MASK) != 0) {
				shape = class_259.method_1084(shape, east);
			}

			shapes[i] = shape;
		}

		return shapes;
	}



	@Override
	@SuppressWarnings("deprecation")
	public class_1269 method_9534(class_2680 state, class_1937 world, class_2338 pos, class_1657 player, class_1268 hand, class_3965 result) {
		class_1799 held = player.method_5998(hand);

		if(stripped != null && held.method_7909() instanceof class_1766) {
			class_1766 tool = (class_1766) held.method_7909();

			if(tool.method_7856(state) || tool.method_7865(held, state) > 1.0F) {
				world.method_8396(player, pos, class_3417.field_14675, class_3419.field_15245, 1.0F, 1.0F);

				if(!world.field_9236) {
					class_2680 target = stripped.get().method_9564()
						.method_11657(BareSmallLogBlock.UP, state.method_11654(BareSmallLogBlock.UP))
						.method_11657(BareSmallLogBlock.DOWN, state.method_11654(BareSmallLogBlock.DOWN))
						.method_11657(BareSmallLogBlock.NORTH, state.method_11654(BareSmallLogBlock.NORTH))
						.method_11657(BareSmallLogBlock.SOUTH, state.method_11654(BareSmallLogBlock.SOUTH))
						.method_11657(BareSmallLogBlock.EAST, state.method_11654(BareSmallLogBlock.EAST))
						.method_11657(BareSmallLogBlock.WEST, state.method_11654(BareSmallLogBlock.WEST))
						.method_11657(BareSmallLogBlock.WATERLOGGED, state.method_11654(BareSmallLogBlock.WATERLOGGED));

					world.method_8501(pos, target);

					held.method_7956(1, player, consumedPlayer -> consumedPlayer.method_20236(hand));
				}

				return class_1269.field_5812;
			}
		}

		return class_1269.field_5814;
	}

	@SuppressWarnings("deprecation")
	@Override
	public boolean method_16362(class_2680 state, class_1922 view, class_2338 pos) {
		return false;
	}

	@SuppressWarnings("deprecation")
	@Override
	public boolean method_9523(class_2680 state, class_1922 view, class_2338 pos, class_1299<?> entityType) {
		return class_2246.field_10620.method_9523(state, view, pos, entityType);
	}

	@Override
	protected void method_9515(class_2689.class_2690<class_2248, class_2680> builder) {
		super.method_9515(builder);

		builder.method_11667(UP, DOWN, NORTH, SOUTH, EAST, WEST, WATERLOGGED);
	}

	private boolean shouldConnectTo(class_2680 state, boolean solid) {
		class_2248 block = state.method_11614();

		return solid || block instanceof BareSmallLogBlock;
	}

	@Override
	public void method_9567(class_1937 world, class_2338 pos, class_2680 state, class_1309 entity, class_1799 stack) {
		super.method_9567(world, pos, state, entity, stack);

		for (class_2350 direction : class_2350.values()) {
			if (world.method_8320(pos.method_10093(direction)).method_11614() instanceof BareSmallLogBlock) {
				world.method_8501(pos.method_10093(direction),
					getNeighborUpdateState(world.method_8320(pos.method_10093(direction)),
						direction.method_10153(),
						world.method_8320(pos),
						world,
						pos.method_10093(direction),
						pos));
			}
		}
	}

	@Override
	public class_2680 method_9605(class_1750 context) {

		class_4538 world = context.method_8045();
		class_2338 pos = context.method_8037();
		class_3610 fluid = context.method_8045().method_8316(context.method_8037());

		if (context.method_8036() == null) {
			return fluid.method_15772().equals(class_3612.field_15910) ? this.method_9564().method_11657(WATERLOGGED, true) : this.method_9564();
		}

		class_2338 upPos = pos.method_10084();
		class_2338 downPos = pos.method_10074();
		class_2338 northPos = pos.method_10095();
		class_2338 eastPos = pos.method_10078();
		class_2338 southPos = pos.method_10072();
		class_2338 westPos = pos.method_10067();

		class_2680 upState = world.method_8320(upPos);
		class_2680 downState = world.method_8320(downPos);
		class_2680 northState = world.method_8320(northPos);
		class_2680 eastState = world.method_8320(eastPos);
		class_2680 southState = world.method_8320(southPos);
		class_2680 westState = world.method_8320(westPos);

		boolean up = this.shouldConnectTo(upState, class_2248.method_20045(upState, world, upPos, class_2350.field_11036));
		boolean down = this.shouldConnectTo(downState, class_2248.method_20045(downState, world, downPos, class_2350.field_11033));
		boolean north = this.shouldConnectTo(northState, class_2248.method_20045(northState, world, northPos, class_2350.field_11035));
		boolean east = this.shouldConnectTo(eastState, class_2248.method_20045(eastState, world, eastPos, class_2350.field_11039));
		boolean south = this.shouldConnectTo(southState, class_2248.method_20045(southState, world, southPos, class_2350.field_11043));
		boolean west = this.shouldConnectTo(westState, class_2248.method_20045(westState, world, westPos, class_2350.field_11034));

		return this.method_9564()
			.method_11657(UP, up)
			.method_11657(DOWN, down)
			.method_11657(NORTH, north)
			.method_11657(EAST, east)
			.method_11657(SOUTH, south)
			.method_11657(WEST, west)
			.method_11657(WATERLOGGED, fluid.method_15772() == class_3612.field_15910);
	}

	@Override
	@SuppressWarnings("deprecation")
	public class_3610 method_9545(class_2680 state) {
		return state.method_11654(WATERLOGGED) ? class_3612.field_15910.method_15729(false) : super.method_9545(state);
	}

	@Override
	public class_2680 method_9598(class_2680 state, class_2470 rotation) {
		switch (rotation) {
			case field_11464:
				return state.method_11657(NORTH, state.method_11654(SOUTH)).method_11657(EAST, state.method_11654(WEST)).method_11657(SOUTH, state.method_11654(NORTH)).method_11657(WEST, state.method_11654(EAST));
			case field_11465:
				return state.method_11657(NORTH, state.method_11654(EAST)).method_11657(EAST, state.method_11654(SOUTH)).method_11657(SOUTH, state.method_11654(WEST)).method_11657(WEST, state.method_11654(NORTH));
			case field_11463:
				return state.method_11657(NORTH, state.method_11654(WEST)).method_11657(EAST, state.method_11654(NORTH)).method_11657(SOUTH, state.method_11654(EAST)).method_11657(WEST, state.method_11654(SOUTH));
			default:
				return state;
		}

	}

	@Override
	@SuppressWarnings("deprecation")
	public class_2680 method_9569(class_2680 state, class_2415 mirror) {
		switch (mirror) {
			case field_11300:
				return state.method_11657(NORTH, state.method_11654(SOUTH)).method_11657(SOUTH, state.method_11654(NORTH));
			case field_11301:
				return state.method_11657(EAST, state.method_11654(WEST)).method_11657(WEST, state.method_11654(EAST));
			default:
				return super.method_9569(state, mirror);
		}
	}

	@Override
	@SuppressWarnings("deprecation")
	public boolean method_9516(class_2680 state, class_1922 view, class_2338 pos, class_10 blockPlacementEnvironment_1) {
		return false;
	}

	public class_2680 getNeighborUpdateState(class_2680 state, class_2350 fromDirection, class_2680 neighbor, class_1936 world, class_2338 pos, class_2338 neighborPos) {
		if (state.method_11654(WATERLOGGED)) {
			world.method_8405().method_8676(pos, class_3612.field_15910, class_3612.field_15910.method_15789(world));
		}

		boolean up = fromDirection == class_2350.field_11036 && this.shouldConnectTo(neighbor, class_2248.method_20045(neighbor, world, neighborPos, class_2350.field_11033)) || state.method_11654(UP);
		boolean down = fromDirection == class_2350.field_11033 && this.shouldConnectTo(neighbor, class_2248.method_20045(neighbor, world, neighborPos, class_2350.field_11036)) || state.method_11654(DOWN);
		boolean north = fromDirection == class_2350.field_11043 && this.shouldConnectTo(neighbor, class_2248.method_20045(neighbor, world, neighborPos, class_2350.field_11035)) || state.method_11654(NORTH);
		boolean east = fromDirection == class_2350.field_11034 && this.shouldConnectTo(neighbor, class_2248.method_20045(neighbor, world, neighborPos, class_2350.field_11039)) || state.method_11654(EAST);
		boolean south = fromDirection == class_2350.field_11035 && this.shouldConnectTo(neighbor, class_2248.method_20045(neighbor, world, neighborPos, class_2350.field_11043)) || state.method_11654(SOUTH);
		boolean west = fromDirection == class_2350.field_11039 && this.shouldConnectTo(neighbor, class_2248.method_20045(neighbor, world, neighborPos, class_2350.field_11034)) || state.method_11654(WEST);

		return state
			.method_11657(UP, up)
			.method_11657(DOWN, down)
			.method_11657(NORTH, north)
			.method_11657(EAST, east)
			.method_11657(SOUTH, south)
			.method_11657(WEST, west);
	}

	@Override
	public boolean method_9579(class_2680 state, class_1922 view, class_2338 pos) {
		return !state.method_11654(WATERLOGGED);
	}

	@Override
	@SuppressWarnings("deprecation")
	public class_265 method_9530(class_2680 state, class_1922 view, class_2338 pos, class_3726 context) {
		return this.boundingShapes[this.getShapeIndex(state)];
	}

	@Override
	@SuppressWarnings("deprecation")
	public class_265 method_9549(class_2680 state, class_1922 view, class_2338 pos, class_3726 context) {
		return this.collisionShapes[this.getShapeIndex(state)];
	}
}
